use serde::Serialize;
use thiserror::Error;

use crate::core::packet::{Packet, PacketError};

// #[allow(unused_variables, dead_code)]
#[derive(Error, Debug)]
pub enum MessageError {
    #[error("{0}")]
    InvalidPacket(#[from] PacketError),
    #[error("{0}")]
    InvalidJsonBody(#[from] serde_json::Error),
    #[error("parse json command failed")]
    InvalidCommand,
    #[error("no message available")]
    NoMessage,
    #[error("unknown message error")]
    Unknown,
}


#[allow(unused_variables, dead_code)]
#[derive(Serialize, Debug)]
#[serde(tag = "cmd", content = "data")]
pub enum Message {
    #[serde(rename = "DAMMU_MSG")]
    DanmuMsg {
        sender: String
    },
    #[serde(rename = "SUPER_CHAT_MESSAGE")]
    SuperChatMessage {
        message: String
    },
    #[serde(rename = "SUPER_CHAT_MESSAGE_DELETE")]
    SuperChatMessageDelete {
        id: u64
    },
    #[serde(rename = "SEND_GIFT")]
    SendGift {
        id: u64
    },
}

impl Message {
    fn parse_packet(single_packet: &mut Packet) -> Result<Message, MessageError> {
        let body = &single_packet.get_body().unwrap()[0];
        let message: serde_json::Result<Message> = serde_json::from_slice(body);
        match message {
            Ok(cmd_json) => Ok(cmd_json),
            // TODO: 对错误类型进行处理，字段不存在和语法错误
            Err(err) => MessageError::InvalidJsonBody(err)
        }
    }

    /// 从 Packet 中提取 Message
    ///
    /// 只有 Json 格式的非业务消息会被提取，
    /// 即心跳包回复，认证请求回复等不会被解析，返回 MessageError::NoMessage
    pub fn from_packet(packet: &mut Packet) -> Result<Vec<Self>, MessageError> {
        let mut packets = packet.get_body_packets()?;
        if packets.len() == 0 {
            return Err(MessageError::NoMessage);
        };

        let messages: Vec<Message> = packets.iter_mut().map(|packet| {
            let parsed = Self::parse_packet(packet);
            // ignore error
            match parsed {
                Ok(msg) => msg,
                Err(_) => ()
            }
        }).collect();
        if messages.len() == 0 {
            Err(MessageError::NoMessage)
        } else {
            Ok(messages)
        }
    }
}

#[cfg(test)]
mod tests {
    #[test]
    fn test_from_packet() {}
}